{- glossary:
	cb = code block, d = descendant, pub = not private definition, li = loop init, ga = gather
	, def = definition, var = variable, expr = expression, decl = declaration -}
{-# LANGUAGE DeriveFoldable, DeriveFunctor #-}

module Language.SsaWhile where
import Data.Function
import Data.Maybe
import Data.Monoid
import Data.List as List
import qualified Data.List.Last as LL
import Data.Fix {- Hackage.data-fix -}
import qualified Data.Foldable as Foldable
import qualified Data.Set as Set {- Hackage.containers -}
import qualified Data.Map.Strict as Map {- Hackage.containers -}
import Data.Graph.Inductive as Graph {- Hackage.fgl -}
import Control.Monad as Monad
import Control.Monad.State (State, state, execState) {- Hackage.mtl -}

type CbName = String
type VarName = String
type VarD = LL.T CbName VarName
data Profit a = Const Int | Add a a | Mult a a | Equal a a
	deriving (Functor, Foldable.Foldable)
data ExprProfit a = Var (VarD) | Profit (Profit a)
	deriving (Functor)
data Expr = Loop (Fix ExprProfit) Expr | ExprProfit (ExprProfit Expr)
type Cb = [Decl]
data Decl = Assign Bool VarName Expr | While CbName (Fix ExprProfit) Cb

exprProfitAlgebraMap :: (a -> [VarD]) -> ExprProfit a -> [VarD]
exprProfitAlgebraMap f x = case x of
	Var v -> [v]
	Profit x -> Foldable.concatMap f x
exprProfitAlgebra :: ExprProfit [VarD] -> [VarD]
exprProfitAlgebra = exprProfitAlgebraMap id
exprProfitCatam :: Fix ExprProfit -> [VarD]
exprProfitCatam = cata exprProfitAlgebra

gaUseIteration, gaUseLoop, gaUseLi :: Expr -> [VarD]
gaUseIteration = fix $ \rc x -> case x of
	Loop _ _ -> []
	ExprProfit x -> exprProfitAlgebraMap rc x
gaUseLoop = fix $ \rc x -> case x of
	Loop _ x -> rc x
	ExprProfit x -> exprProfitAlgebraMap rc x
gaUseLi = fix $ \rc x -> case x of
	Loop x _ -> exprProfitCatam x
	ExprProfit x -> case x of
		Var _ -> []
		Profit x -> Foldable.concatMap rc x

gaAssignEdges :: (Expr -> t) -> Cb -> [(VarD, t)]
gaAssignEdges gaUse = catMaybes . fmap (\x -> case x of
	Assign _ v e -> Just (LL.Last v, gaUse e)
	While cbName e cb -> Nothing
	)

data Error = DupDef VarName | DupCbName CbName | VarChildInLi VarD | DefUseCycle [[VarD]] | NoVarChild VarD
	deriving (Show)

errorPath cbName (Left x) = Left (LL.Cons cbName x)
errorPath _ x = x
errorHere = Left . LL.Last

getDef (Assign _ v _) = Just v
getDef _ = Nothing
getPub (Assign False v _) = Just v
getPub _ = Nothing
getCbName (While cbName _ _) = Just cbName
getCbName _ = Nothing
gaDeclName getName = fix (\rc s xs -> case xs of
	[] -> Right s
	x:xs -> case getName x of
		Nothing -> rc s xs
		Just n -> case n `Set.member` s of
			True -> Left n
			False -> rc (Set.insert n s) xs
	) Set.empty
gaDefAndCbName cb = case gaDeclName getCbName cb of
	Left x -> errorHere (DupCbName x)
	Right _ -> case gaDeclName getDef cb of
		Left x -> errorHere (DupDef x)
		Right defs -> case gaDeclName getPub cb of
			Left x -> undefined
			Right pubs -> Right (defs, pubs)

getVarChild (LL.Last _) = Nothing
getVarChild x = Just x

type VarEdges = [(VarD, [VarD])]

gaLiEdges :: Cb -> Either (LL.T a Error) VarEdges
gaLiEdges cb = let
	liEdges = gaAssignEdges gaUseLi cb
	in case (getFirst . mconcat . fmap (First . getVarChild)) (liEdges >>= snd) of
		Just x -> errorHere (VarChildInLi x)
		Nothing -> Right liEdges

checkVarChildIsDefined uses defs = let
	lookupWhileDef x = case x `Set.member` (Set.fromList defs) of
		False -> Just x
		True -> Nothing
	in case (getFirst . mconcat . fmap (First . (>>= lookupWhileDef) . getVarChild)) uses of
		Nothing -> Right ()
		Just x -> errorHere (NoVarChild x)

type StateGr a = State (Node, Map.Map a Node, Gr a ())

insertNodeVar :: (Ord a) => a -> StateGr a Node
insertNodeVar x = state $ \(k, m, g) -> case Map.insertLookupWithKey (\_ _ k -> k) x k m of
	(existed, m1) -> case existed of
		Nothing -> (k, (k+1, m1, insNode (k, x) g))
		Just k1 -> (k1, (k, m1, g))

insertSuccessors :: (Ord a) => (a, [a]) -> StateGr a ()
insertSuccessors (x, ys) = do
	xg <- insertNodeVar x
	Monad.forM_ ys $ \y -> do
		yg <- insertNodeVar y
		state $ \(k, m, g) -> ((), (k, m, insEdge (xg, yg, ()) g))

createGraph :: VarEdges -> Gr VarD ()
createGraph edgeList = let
	s0 = (0, Map.empty, Graph.empty)
	c = Monad.mapM_ insertSuccessors edgeList
	in case execState c s0 of
		(_, m, g) -> g

getCycleScc :: Gr a () -> [[a]]
getCycleScc g = let
	f xs = case xs of
		[x] -> x `List.elem` suc g x
		_ -> True
	in (fmap (fmap (fromJust . lab g)) . List.filter f) (Graph.scc g)

graphCompLi :: (Set.Set VarName, Set.Set VarName)
	-> VarEdges -> Gr VarD () -> VarEdges
graphCompLi (defs, pubs) liEdges g = let
	isExported tops x = case x of
		LL.Last v -> v `Set.member` tops
		_ -> True
	lab1 = fromJust . lab g
	dep x = (lab1 x, let
		suc1 = fmap lab1 (suc g x)
		sucLi = do
			x <- suc1
			case Map.lookup x (Map.fromList liEdges) of
				Nothing -> []
				Just x -> x
		in (List.filter (not . isExported defs)) suc1 ++ sucLi
		)
	in (fmap dep . List.filter (isExported pubs . lab1) . Graph.nodes) g

gaCbEdges :: [VarD] -> Cb -> Either (LL.T CbName Error) VarEdges
gaCbEdges closeVars cb = do
	defPubs <- gaDefAndCbName cb
	liEdges <- gaLiEdges cb
	whileEdges <- gaWhileEdges cb
	let assignLoopEdges = gaAssignEdges ((closeVars++) . gaUseLoop) cb
	checkVarChildIsDefined
		(closeVars ++ (assignLoopEdges >>= snd))
		(fmap fst whileEdges)
	(case (getCycleScc . createGraph) (gaAssignEdges gaUseIteration cb ++ whileEdges) of
		[] -> Right ()
		xs -> errorHere (DefUseCycle xs)
		)
	return ((graphCompLi defPubs liEdges . Graph.trc . createGraph)
		(assignLoopEdges ++ whileEdges))

gaWhileEdges :: Cb -> Either (LL.T CbName Error) VarEdges
gaWhileEdges = fix (\rc ys xs -> case xs of
	[] -> Right ys
	x:xs -> case x of
		Assign _ _ _ -> rc ys xs
		While cbName e cb -> errorPath cbName $ do
			cbEdges <- gaCbEdges (exprProfitCatam e) cb
			let
				g (v, vs) = (LL.Cons cbName v, vs)
				in rc (fmap g cbEdges ++ ys) xs
	) []

